//@author: fibo
//@help: 
//@tags: AttractorCenter, 3d, cube
//@credits: bo27

// -----------------------------------------------------------------------------
// PARAMETERS:
// -----------------------------------------------------------------------------

//transforms
float4x4 tW: WORLD;        //the models world matrix
float4x4 tV: VIEW;         //view matrix as set via Renderer (DX9)
float4x4 tWV: WORLDVIEW;
float4x4 tWVP: WORLDVIEWPROJECTION;
float4x4 tP: PROJECTION;   //projection matrix as set via Renderer (DX9)


//light properties
float3 lPos <string uiname="Light Position";> = {0, 5, -2};       //light position in world space
float lAtt0 <String uiname="Light Attenuation 0"; float uimin=0.0;> = 0;
float lAtt1 <String uiname="Light Attenuation 1"; float uimin=0.0;> = 0.3;
float lAtt2 <String uiname="Light Attenuation 2"; float uimin=0.0;> = 0;
float4 lAmb  : COLOR <String uiname="Ambient Color";>  = {0.15, 0.15, 0.15, 1};
float4 lDiff : COLOR <String uiname="Diffuse Color";>  = {0.85, 0.85, 0.85, 1};
float4 lSpec : COLOR <String uiname="Specular Color";> = {0.35, 0.35, 0.35, 1};
float lPower <String uiname="Power"; float uimin=0.0;> = 25.0;     //shininess of specular highlight
float lRange <String uiname="Light Range"; float uimin=0.0;> = 10.0;

//phong point function
float4 PhongPoint(float3 PosW, float3 NormV, float3 ViewDirV, float3 LightDirV)
{

    float d = distance(PosW, lPos);
    float atten = 0;

    //compute attenuation only if vertex within lightrange
    if (d<lRange)
    {
       atten = 1/(saturate(lAtt0) + saturate(lAtt1) * d + saturate(lAtt2) * pow(d, 2));
    }

    float4 amb = lAmb * atten;
    amb.a = 1;

    //halfvector
    float3 H = normalize(ViewDirV + LightDirV);

    //compute blinn lighting
    float4 shades = lit(dot(NormV, LightDirV), dot(NormV, H), lPower);

    float4 diff = lDiff * shades.y * atten;
    diff.a = 1;

    //reflection vector (view space)
    float3 R = normalize(2 * dot(NormV, LightDirV) * NormV - LightDirV);

    //normalized view direction (view space)
    float3 V = normalize(ViewDirV);

    //calculate specular light
    float4 spec = pow(max(dot(R, V),0), lPower*.2) * lSpec;

    return (amb + diff) + spec;
}

float3 AttractorCenter;
float AttractorWidth <String uiname="Attractor Width"; float uimin=0.0;> = 1;
float AttractorHeight <String uiname="Attractor Height"; float uimin=0.0;> = 1;
float AttractorDepth <String uiname="Attractor Depth"; float uimin=0.0;> = 1;
float AttractorPower <String uiname="Attractor Power";> = 0;
float AttractorStrength <String uiname="Attractor Strength";> = 0;
float3 AttractorRotate;

bool isInside(  
  float3 v, // vertex
  float3 c, // center
  float3 rot, // rotate
  float  w, // width
  float  h, // height
  float  d  // depth
){
  // i need to generate from scratch rotation matrices
  // keeping compatibility with Rotate node.
  float TWOPI = 6.2832;
  rot *= TWOPI;
  float4x4 rotX = {
    {1,          0,         0,0},
    {0, cos(rot.x),sin(rot.x),0},
    {0,-sin(rot.x),cos(rot.x),0},
    {0,          0,         0,1}
  };
  float4x4 rotY = {
    {cos(rot.y),0,-sin(rot.y),0},
    {0         ,1,          0,0},
    {sin(rot.y),0, cos(rot.y),0},
    {0,         0,          0,1}
  };
  float4x4 rotZ = {
    {cos(rot.z),-sin(rot.z),0,0},
    {sin(rot.z), cos(rot.z),0,0},
    {         0,          0,1,0},
    {         0,          0,0,1}
  };

  // here it is, now i translate the vertex to the origin ...
  v = v - c;
  // ... i apply the rotations ...
  v = mul(v, rotX);
  v = mul(v, rotY);
  v = mul(v, rotZ);
  // ... and i move back the vertex.
  v = v + c;
  
  if( (abs(v.x-c.x)>w/2) ||
      (abs(v.y-c.y)>h/2) ||
      (abs(v.z-c.z)>d/2)
  ){ return false; } else { return true; }
}

float3 attractor3dCube (
  float3 v, // vertex
  float3 c, // center
  float3 rot, // rotate
  float  p, // power
  float  s, // strenght
  float  w, // width
  float  h, // height
  float  d  // depth
){

  float r = max(w, h);
  r = max(r, d);
  
  if(isInside(v,c,rot,w,h,d))
  {

    v += (v-c)*s*(pow((length(v-c)/r),p)*sign(length(v-c)/r)/(length(v-c)/r)-1);

  }
  return v;
}


//texture
texture Tex <string uiname="Texture";>;
sampler Samp = sampler_state    //sampler for doing the texture-lookup
{
    Texture   = (Tex);          //apply a texture to the sampler
    MipFilter = LINEAR;         //set the sampler states
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

float4x4 tTex: TEXTUREMATRIX <string uiname="Texture Transform";>;
float4x4 tColor <string uiname="Color Transform";>;

struct vs2ps
{
    float4 PosWVP: POSITION;
    float4 TexCd : TEXCOORD0;
    float3 LightDirV: TEXCOORD1;
    float3 NormV: TEXCOORD2;
    float3 ViewDirV: TEXCOORD3;
    float3 PosW: TEXCOORD4;
};

// -----------------------------------------------------------------------------
// VERTEXSHADERS
// -----------------------------------------------------------------------------

vs2ps VS(
    float4 PosO: POSITION,
    float3 NormO: NORMAL,
    float4 TexCd : TEXCOORD0)
{
    //inititalize all fields of output struct with 0
    vs2ps Out = (vs2ps)0;

    Out.PosW = mul(PosO, tW);

    //inverse light direction in view space
    float3 LightDirW = normalize(lPos - Out.PosW);
    Out.LightDirV = mul(LightDirW, tV);
    
    //normal in view space
    Out.NormV = normalize(mul(NormO, tWV));

    PosO = mul(PosO, tW);
    PosO.xyz = attractor3dCube(PosO.xyz,AttractorCenter,AttractorRotate,AttractorPower,AttractorStrength,AttractorWidth,AttractorHeight,AttractorDepth);
    PosO = mul(PosO, tV);
    
    //position (projected)
    //Out.PosWVP  = mul(PosO, tWVP);
    Out.PosWVP = mul(PosO, tP);
    Out.TexCd = mul(TexCd, tTex);
    Out.ViewDirV = -normalize(mul(PosO, tWV));
    return Out;
}

// -----------------------------------------------------------------------------
// PIXELSHADERS:
// -----------------------------------------------------------------------------

float foo(float bar)
{
    return bar;
}

float Alpha <float uimin=0.0; float uimax=1.0;> = 1;

float4 PS(vs2ps In): COLOR
{
    //In.TexCd = In.TexCd / In.TexCd.w; // for perpective texture projections (e.g. shadow maps) ps_2_0
    
    float4 col = tex2D(Samp, In.TexCd);

    col.rgb *= PhongPoint(In.PosW, In.NormV, In.ViewDirV, In.LightDirV);
    col.a *= Alpha;

    return mul(col, tColor);
}


// -----------------------------------------------------------------------------
// TECHNIQUES:
// -----------------------------------------------------------------------------

technique TPhongPoint
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_1_1 VS();
        PixelShader = compile ps_2_0 PS();
    }
}

technique TFallbackGouraudPointFF
{
    pass P0
    {
        //transformations
        WorldTransform[0]   = (tW);
        ViewTransform       = (tV);
        ProjectionTransform = (tP);

        //material
        MaterialAmbient  = {1, 1, 1, 1};
        MaterialDiffuse  = {1, 1, 1, 1};
        MaterialSpecular = {1, 1, 1, 1};
        MaterialPower    = (lPower);

        //texturing
        Sampler[0] = (Samp);
        TextureTransform[0] = (tTex);
        TexCoordIndex[0] = 0;
        TextureTransformFlags[0] = COUNT2;
        //Wrap0 = U;  // useful when mesh is round like a sphere

        //lighting
        LightEnable[0] = TRUE;
        Lighting       = TRUE;
        SpecularEnable = TRUE;
        
        LightType[0]     = POINT;
        LightAmbient[0]  = (lAmb);
        LightDiffuse[0]  = (lDiff);
        LightSpecular[0] = (lSpec);
        LightPosition[0] = (lPos);
        LightRange[0]    = (lRange);
        LightAttenuation0[0] = (lAtt0);
        LightAttenuation1[0] = (lAtt1);
        LightAttenuation2[0] = (lAtt2);

        //shading
        ShadeMode = GOURAUD;
        VertexShader = NULL;
        PixelShader  = NULL;
    }
}
